/*
jpg_to_mjpegtools_yuv:  converts jpg images to mjpeg tools YUV format.

jpg_to_mjpegtools_yuv is registered Copyright (C) 2008 and later <Jan Panteltje>
email: panteltje@yahoo.com

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


                            /* set TABS to 4 to read this */

#include "jpg_to_mjpegtools_yuv.h"


char *strsave(char *s) /* save char array s somewhere */
{
char *p;

p = malloc(strlen(s) +  1);
if(p) strcpy(p, s);

return p;
} /* end function strsave */


int print_usage()
{
fprintf(stderr, "\
Usage:\n\
jpg_to_mjpegtools_yuv [-b] [-f start_number] [-h] [-l stop_number] [-q] [-s seconds_per_frame]\n\n\
-b                   no color.\n\
-d                   directory where to look for jpg pictures, default current dir.\n\
-f int               start number of jpg, default 1.\n\
-h                   help (this help).\n\
-l int               end number of jpg, default infinity.\n\
-m                   manual mode, type ENTER for next picture.\n\
-q                   quit, no messages to stderr.\n\
-s int               seconds per frame, default %d.\n\
-v                   debug, prints functions and arguments.\n\
-x                   x11 display.\n\
",\
DEFAULT_SECONDS_PER_FRAME);

return 1;
} /* end function print_usage */


int main(int argc, char **argv)
{
int a, c;
char *ptr;
FILE *fptr;
char *window_border_info[1];
struct stat *statptr;
int last_jpg;
char temp[80];
char *directory_name;

/* set defaults */
debug_flag = 0;
quiet_flag = 0;
no_color_flag = 0;
seconds_per_frame = DEFAULT_SECONDS_PER_FRAME;
x_display = 0;
manual_flag = 0;
jpg_number = 1;
last_jpg = -1;
directory_name = strsave("");
if(! directory_name)
	{
	fprintf(stderr, "jpg_to_mjpegtools_yuv: could not allocate space for directory name, aborting.\n");

	exit(1);
	}
/* end defaults */

/*select con io 1 byte at the time*/
//setbuf(stdout, NULL);
//setbuf(stdin, NULL);

/* get command line options */
while(1)
	{
	a = getopt(argc, argv, "bd:f:hl:mqs:vx");
	if(a == -1) break;

	switch(a)
		{
		case 'b': // force BW
			no_color_flag = 1;
			break;
		case 'd': // directory_name for jpg
			directory_name = strsave(optarg);
			if(! directory_name)
				{
				fprintf(stderr, "jpg_to_mjpegtools_yuv: could not allocate space for directory name, aborting.\n");
				
				exit(1);
				}
			break;
		case 'f': // start number jpg
			jpg_number = atoi(optarg);
			break;
		case 'h': // help
			print_usage();
			exit(1);
			break;
		case 'l': // end number jpg
			last_jpg = atoi(optarg);
			break;
		case 'm': // manual mode
			manual_flag = 1;
			break;
		case 'q': // quiet mode
			quiet_flag = 1;
			break;
		case 's': // set second per frames for YUV output (mplayer will display in that speed).
			seconds_per_frame = atoi(optarg);
			break;
		case 'v': // verbose / debug
			debug_flag = 1;
			break;
		case 'x': // x display
			x_display = 1;
			break;
		case -1:
        	break;
		case '?':
			if (isprint(optopt) )
 				{
 				fprintf(stderr, "jpg_to_mjpegtools_yuv: unknown option `-%c'.\n", optopt);
 				}
			else
				{
				fprintf(stderr, "jpg_to_mjpegtools_yuv: uUnknown option character `\\x%x'.\n", optopt);
				}
			print_usage();

			exit(1);
			break;			
		default:
			print_usage();

			exit(1);
			break;
		}/* end switch a */
	}/* end while getopt() */

/* ID */
if(! quiet_flag)
	{
	fprintf(stderr, "jpg_to_mjpegtools_yuv-%s copyright Jan Panteltje 21st century.\n", PVERSION);
	}

statptr = (struct stat*) malloc(sizeof(struct stat) );
if(! statptr)
	{
	fprintf(stderr, "jpg_to_mjpegtools_yuv: could not allocate space for statptr, aborting.\n");

	exit(1);
	}

/*
This value should be smaller then a normally received jpeg.... but we do not know the minimum really.
Theoretically if the jpgs are all much smaller then this, buffer problems in read could result.
*/
while(1)/* main communications loop */
	{
	/* read in next jpg file, fill jpeg_buffer, set content_length */

//	fptr = fopen("/root/berlepschs-six-wired-bird-of-paradise-109394.jpg", "r");

	if( strcmp(directory_name, "") == 0)
		{
		sprintf(temp, "%d.jpg", jpg_number);
		}
	else
		{
		sprintf(temp, "%s/%d.jpg", directory_name, jpg_number);
		}

	if(debug_flag)
		{
		fprintf(stderr, "jpg_to_mjpegtools_yuv: opening %s for read\n", temp);
		}

	fptr = fopen(temp, "r");
	if(! fptr)
		{
		fprintf(stderr, "jpg_to_mjpegtools_yuv: could not open file %s for read, aborting.\n", temp);
		
		exit(1);
		}

	/* get the file size */
//	int fstat (int FILEDES, struct stat *BUF)
	if(fstat(fileno(fptr), statptr) == -1)
		{
		fprintf(stderr, "jpg_to_mjpegtools_yuv: fstat() failed, aborting.\n");

		exit(1);		
		}

	if(debug_flag)
		{
		fprintf(stderr, "jpg_to_mjpegtools_yuv: statptr->st_size=%d\n", (int)statptr -> st_size);
		}

	jpeg_buffer = (char *)malloc(statptr -> st_size + 1);
	if(! jpeg_buffer)
		{
		fprintf(stderr, "jpg_to_mjpegtools_yuv: could not allocate space for jpeg_buffer, aborting.\n");

		exit(1);
		}

	content_length = 0;
	ptr = jpeg_buffer;
	while(1)
		{
		c = fgetc(fptr);
		if(c == EOF)
			{
			fclose(fptr);

			break;
			}
			
		*ptr = c;
		ptr++;
		content_length++;
		}

	/* process the jpeg buffer */
	if(! process_jpeg_buffer(1, window_border_info) )
		{
		fprintf(stderr, "jpg_to_mjpegtools_yuv: process_jpeg_buffer() failed, aborting.\n");
						
		exit(1);
		}

	if(last_jpg != -1)
		{
		if(jpg_number == last_jpg)
			{
			if(! quiet_flag)
				{
				fprintf(stderr, "jpg_to_mjpegtools_yuv: stopped at %d as requested on command line.\n", last_jpg); 
				}
				
			exit(0);			
			} 
		}
		
	jpg_number++;

	if(debug_flag)
		{
		fprintf(stderr, "jpg_to_mjpegtools_yuv: loop\n");
		}
	} /* end main communications loop */

exit(0);
} /* end function main */


